"""MsxClient class for connecting to remote instance.

Classes
    MsxClient

Functions
    handle_unknown_response(msg: Optional[str] = None) -> ApiError
"""
import logging
from typing import Optional

import requests
from dacite import from_dict

import pymsx.handlers as handle
from pymsx.api.commands import Commands

# api classes
from pymsx.api.datasets import Datasets
from pymsx.config import Configuration, app_config
from pymsx.exceptions import ApiResponseError, InvalidTokenError
from pymsx.schemas import ApiError, ApiMessage, Credentials, HealthStatus, TokenResponse

logger = logging.getLogger(__name__)


class MsxClient:
    """
    Main class for access to remote msx functionality.

    Note:
        An email, password, or token is required to connect. If passed in, they will
        be used. If not, environment variables will be checked. If all these fails,
        the construction of this class will fail.

    Args:
        email (str, optional): email used to connect
        password (str, optional): password used to connect
        token (str, optional): the token used to connect

    Attributes:
        validated (bool): whether the token stored is valid
        base_url (str): the url the client is connecting to
        org_id (str): id of the organization currently connected
        email (str, optional): email if provided
        password (str, optional): password if provided
        token (str): the token being used to communicate to the remote msx server

        datasets (:obj: `Datasets`): Datasets related commands
        commands (:obj: `Commands`): Commands and task requests.
    """

    validated: bool = False
    base_url: str
    org_id: str

    def __init__(
        self,
        email: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
    ):
        """Create a connection to org's remote msx instance."""
        config = self.config

        self.base_url = config.base_url
        self.token = token or config.token
        self.email = email or config.email
        self.password = password or config.password

        if self.token is None and self.email is None and self.password is None:
            raise ValueError("Either a token or email/password is required.")

        try:
            _ = self.connect()
            logger.info(f"Sucessfully connected to msx server {self.org_id}")
            self.validated = True
        except ApiResponseError as e:
            logger.error(f"Error connecting: {e.error}")
            if "You are not authorized to make this request" in e.error:
                raise InvalidTokenError
            else:
                raise e

        self.datasets = Datasets(self)
        self.commands = Commands(self)

    @property
    def config(self) -> Configuration:
        """Get runtime application config."""
        return app_config()

    # Internal helpers

    def __get_creds(self) -> Credentials:
        creds = {"email": self.email, "password": self.password}
        return from_dict(data=creds, data_class=Credentials)

    # TODO: Check for exceptions
    def get_token(self) -> TokenResponse:
        """Request new token from api server."""
        if self.email is None or self.password is None:
            raise AttributeError("email and password required to retrieve token.")

        url = f"{self.base_url}/token"
        headers = self.get_auth_headers(with_json=True, with_token=False)
        creds = self.__get_creds()

        logger.debug(
            f"Requesting token using url={url}, " "headers={headers}, creds={creds}"
        )

        res = requests.post(url, json=creds.dict(), headers=headers)
        response = handle.handle_response(res)

        logger.debug("Recieved token response: ", response)

        if isinstance(response, dict):
            token_response: TokenResponse = from_dict(
                data=res.json(), data_class=TokenResponse
            )

            self.token = token_response.token
            self.org_id = token_response.orgId

            return token_response
        elif isinstance(response, ApiError):
            raise ApiResponseError(error=response)
        else:
            raise ApiResponseError()

    def get_auth_headers(
        self, with_json: bool = False, with_token: bool = True
    ) -> dict[str, str]:
        """Get auth headers."""
        headers: dict[str, str] = {}
        if with_token:
            headers = {**headers, "Authorization": f"Bearer {self.token}"}

        if with_json:
            headers = {**headers, "Content-Type": "application/json"}

        return headers

    # Authentication & validation

    def add_org_header(self, headers: Optional[dict[str, str]]) -> dict[str, str]:
        """Add custom org header to requests."""
        org_header = self.config.org_header
        return {**(headers or {}), org_header: self.org_id}

    def validate_token(self) -> ApiMessage:
        """Validate token with remote server."""
        headers = self.get_auth_headers(with_json=False)
        url = f"{self.base_url}/validate"
        req = requests.get(url, headers=headers)
        res = req.json()

        if "message" in res:
            # success message
            return from_dict(data=res, data_class=ApiMessage)
        elif "error" in res:
            # error message
            error = res["error"]
            logger.error(f"Error validating: {error}")
            if "You are not authorized to make this request" in res["error"]:
                raise InvalidTokenError
            else:
                raise ApiResponseError(error=from_dict(data=res, data_class=ApiError))
        else:
            # unknown
            raise ApiResponseError(error=handle.handle_unknown_response())

    def connect(self) -> ApiMessage:
        """Connect to msx using supplied credentials."""
        if self.token is None or self.token == "":
            # try and retrieve an api token
            logger.debug("Fetching token in connect.")
            self.get_token()

        validated = self.validate_token()

        return validated

    # Api calls

    def health(self) -> HealthStatus:
        """Query the current health of the server."""
        url = f"{self.base_url}/health"
        headers = self.get_auth_headers()
        headers = self.add_org_header(headers=headers)

        logger.debug(f"Api request using headers={headers}")

        req = requests.get(url, headers=headers)
        res = req.json()

        if "error" in res:
            raise ApiResponseError(error=from_dict(data=res, data_class=ApiError))
        else:
            return from_dict(data=res, data_class=HealthStatus)



"""Datasets Class used to access remote datasets and metadata.

Datasets represent remote data and their versions. Data is organized by `dataset name`
and versions. Versions are created automatically, and associating versions is done by
reusing existing `dataset names`.

For example, if uploading file `nlp_train.csv` today, the msx server will store that
dataset under the dataset name `nlp_train`. If I then another, newer file also named
`nlp_train.csv`, the msx server will automatically associate them, and store the second
dataset as `version 2`.

The same applies to dataframes, passed in with a name, instead of a file path.

Classes:
    Datasets
"""
import logging
from typing import TYPE_CHECKING, Any, Optional, Union

import requests

from pymsx import handlers
from pymsx.schemas import ApiError, CommandArgs

if TYPE_CHECKING:
    from pymsx.client import MsxClient


logger = logging.getLogger(__name__)


class Commands:
    """
    Main class for issuing commands to remote msx servers.

    ...

    Args:
        client (:obj: `MsxClient`): the client used to perform remote api requests

    Attributes:
        client (:obj: `MsxClient`): the client used to perform remote api requests
    """

    def __init__(self, client: "MsxClient"):
        """Create a new Commands class."""
        self.client = client

    def run(
        self, task: str, args: Optional[list[str]] = None
    ) -> Union[str, Any, ApiError]:
        """
        Execute single command with task and args.

        Args:
            task (str): task to issue to the remote msx server.
            df (:ob: `pandas.DataFrame`): A pandas dataframe
            args (list[str], optional): The args for the task command.

        Returns:
            Any
        """
        args = args or []

        url = f"{self.client.base_url}/msx/"

        auth_headers = self.client.get_auth_headers(with_json=True)
        auth_headers = self.client.add_org_header(headers=auth_headers)

        args = [task, *args]
        cmd_args = CommandArgs(args=args)

        res = requests.post(url, json=cmd_args.dict(), headers={**auth_headers})

        logger.debug(f"Msx command response: {res}")

        response = handlers.handle_response(res)
        return response



"""Application wide configuration."""
import os
import tempfile
from dataclasses import asdict, dataclass
from typing import Optional

import dataconf

env_prefix = "MSX_"


@dataclass
class Configuration:
    """Aggregate app configuration."""

    org_id: Optional[str]
    email: Optional[str]
    password: Optional[str]
    token: Optional[str]

    # dataset specific
    temp_dir: str
    allowed_read_exts: list[str]
    min_stream_size: int

    base_url: str

    org_header: str

    dict = asdict


configuration_defaults = {
    "temp_dir": tempfile.gettempdir(),
    "base_url": "https://api.mosaics.ai",
    "org_header": "X-Org-Id",
    "email": os.getenv("MSX_EMAIL"),
    "password": os.getenv("MSX_PASSWORD"),
    "allowed_read_exts": [".csv", ".parquet", ".json"],
    "min_stream_size": 1024 * 1024 * 1024,
}


def app_config(overrides: Optional[dict] = None) -> Configuration:
    """Runtime application config."""
    config: Configuration = (
        dataconf.multi.dict(configuration_defaults)
        .env(env_prefix)
        .dict(overrides or {})
        .on(Configuration)
    )
    return config



"""Datasets Class used to access remote datasets and metadata.

Datasets represent remote data and their versions. Data is organized by `dataset name`
and versions. Versions are created automatically, and associating versions is done by
reusing existing `dataset names`.

For example, if uploading file `nlp_train.csv` today, the msx server will store that
dataset under the dataset name `nlp_train`. If I then another, newer file also named
`nlp_train.csv`, the msx server will automatically associate them, and store the second
dataset as `version 2`.

The same applies to dataframes, passed in with a name, instead of a file path.

Classes:
    Datasets
"""
import io
import logging
import pathlib
from typing import TYPE_CHECKING, Any, Callable, Optional

import pandas as pd
import requests
from requests_toolbelt.multipart.encoder import (
    MultipartEncoder,
    MultipartEncoderMonitor,
)

if TYPE_CHECKING:
    from pymsx.client import MsxClient


logger = logging.getLogger(__name__)


class Datasets:
    """
    Main class for managing datasets on remote msx servers.

    ...

    Args:
        client (:obj: `MsxClient`): the client used to perform remote api requests

    Attributes:
        client (:obj: `MsxClient`): the client used to perform remote api requests
    """

    def __init__(self, client: "MsxClient"):
        """Create a new Datasets class."""
        self.client = client

    def __get_df_size(self, df: pd.DataFrame):
        try:
            return df.memory_usage(index=True).sum()
        except Exception:
            # size is not essential
            return None

    def __get_mime_type(self, ext: str = ".csv"):
        if ext == ".csv":
            return "text/csv"
        elif ext == ".parquet":
            return "application/parquet"
        elif ext == ".json":
            return "application/parquet"
        else:
            return "text/plain"

    def __get_pd_read_func(self, ext: str = ".csv"):
        read_func = pd.read_csv
        if ext == ".csv":
            pass
        elif ext == ".parquet":
            read_func = pd.read_parquet
        elif ext == ".json":
            read_func = pd.read_json
        else:
            # must be unreachable
            pass
        return read_func

    def __convert_df_to_bytes(self, df: pd.DataFrame, ext: str = ".csv") -> io.BytesIO:
        if ext == ".csv":
            data = io.BytesIO(df.to_csv(index=False).encode("utf-8"))
        elif ext == ".parquet":
            data = io.BytesIO(df.to_parquet(index=False))
        elif ext == ".json":
            data = io.BytesIO(df.to_json(index=False).encode("utf-8"))
        else:
            # must be unreachable
            data = io.BytesIO(df.to_csv(index=False).encode("utf-8"))
        data.seek(0)
        return data

    def add(
        self,
        path_or_name: str,
        df: Optional[pd.DataFrame] = None,
        target: Optional[str] = None,
        store_s3: bool = False,
        df_read_args: Optional[dict[str, Any]] = None,
        callback: Optional[Callable[[MultipartEncoderMonitor], None]] = None,
        **kwargs,
    ):
        """
        Add a dataset to the connected msx server.

        If df is None, then first arg is used as `dataset name` otherwise it must be
        a path to a dataset (file on disk, or soon a location that pandas can parse,
        such as s3)

        Args:
            path_or_name (str): Either a dataset name (if df is provided), or the path
                to a file on disk.
            df (:ob: `pandas.DataFrame`): A pandas dataframe
            target (str, optional): The target (column) of the data that will be used
                when training. If no target is provided, then the last column will be
                used.
            store_s3 (bool, optional): Data can be stored in the isolated msx
                environment, or it can be stored in an accessible (secure) S3 bucket
                that every msx server includes.
            df_read_args (dict[str, Any], optional): If df is not defined, then
                optionally pass in pandas read_* kwargs.
            **kwargs: If kwargs are provided, they will be serialized to dict[str, str]
                and passed to the upload server as is. This is useful because it allows
                passing additional fields to any pipelines or triggers configured to
                run after upload

        Returns:
            {
                path: str,
                **kwargs
            }
        """
        filename = path_or_name

        # READ
        if df is None:
            # attempt reading path_or_name as path
            path = pathlib.Path(path_or_name)
            filename = path.name
            path_ext = path.suffix

            allowed_ext = self.client.config.allowed_read_exts

            if path_ext not in allowed_ext:
                raise ValueError(f"Could not read path type {path_ext}")

            if path_ext == "":
                raise ValueError("Path extension could not be determined.")

            # read path
            read_func = self.__get_pd_read_func(ext=path_ext)

            if df_read_args is not None:
                df = read_func(path_or_name, **df_read_args)
            else:
                df = read_func(path_or_name)

        # WRITE

        # for now using `.csv` for everything write related
        write_ext = ".csv"

        callback = callback or default_monitor

        target = target or "default"
        extra = {}
        if kwargs is not None:
            extra = {
                k: (str(v), io.BytesIO(bytes(str(v), "utf-8")), "text/plain")
                for k, v in kwargs.items()
            }

        filename = f"/s3/{filename}" if store_s3 else f"/datasets/{filename}"

        data = self.__convert_df_to_bytes(df, ext=write_ext)

        # stream = StreamingIterator(size, data)

        e = MultipartEncoder(
            fields={
                **extra,
                "file": (filename, data, self.__get_mime_type(write_ext)),
                "target": (target, io.BytesIO(bytes(target, "utf-8")), "text/plain"),
            }
        )
        m = MultipartEncoderMonitor(e, callback)

        url = f"{self.client.base_url}/upload"

        auth_headers = self.client.get_auth_headers()
        auth_headers = self.client.add_org_header(headers=auth_headers)

        res = requests.post(
            url, data=m, headers={**auth_headers, "Content-type": e.content_type}
        )

        return res.json()


def default_monitor(monitor: MultipartEncoderMonitor) -> None:
    """Monitor for MultipartEncodeMonitor."""
    logger.debug(f"Bytes read: {monitor.bytes_read}")



"""Simple handlers to manage responses from server.

Functions
    handle_response(res: requests.Response) -> Any | str | ApiError
    handle_unknown_response(msg: Optional[str] = None) -> ApiError
"""
import logging
from typing import Any, Optional, Union

import requests
from dacite import from_dict

from pymsx.schemas import ApiError

logger = logging.getLogger(__name__)


def handle_response(res: requests.Response) -> Union[str, Any, ApiError]:
    """Handle server responses."""
    if res.ok:
        try:
            return res.json()
        except requests.exceptions.JSONDecodeError:
            return res.text
    else:
        try:
            return ApiError(error=f"(Code: {res.status_code}): {res.reason}")
        except Exception:
            return ApiError(error="(Code: 0): unknown")


def handle_unknown_response(msg: Optional[str] = None) -> ApiError:
    """Hanlde unknown response types."""
    prefix = "Unknown response"
    message = prefix if msg is None else f"{prefix}: {msg}"
    return from_dict(data={"error": message}, data_class=ApiError)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pymsx', 'pymsx.api']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.26.74,<2.0.0',
 'dacite>=1.8.0,<2.0.0',
 'dataconf>=2.1.3,<3.0.0',
 'numpy>=1.24.2,<2.0.0',
 'pandas>=1.5.3,<2.0.0',
 'pydantic>=1.10.5,<2.0.0',
 'requests-toolbelt>=0.10.1,<0.11.0',
 'requests>=2.28.2,<3.0.0',
 'urllib3==1.25.11']

setup_kwargs = {
    'name': 'pymsx',
    'version': '0.3.0',
    'description': 'Mosaics AI Python Client',
    'long_description': '# 🧰  pymsx - Mosaics AI MSX Client for Python\n\n[![PyPI](https://img.shields.io/pypi/v/pymsx?style=flat-square)](https://pypi.org/project/pymsx/)\n[![Interrogate](https://raw.githubusercontent.com/Mosaics-ai/pymsx/main/assets/interrogate_badge.svg)](https://github.com/Mosaics-AI/pymsx)\n\nThis repository contains the source code for Mosaics AI\'s official python client. This client is currently in pre-alpha version, so feedback is always welcome!\n\nYou are welcome to file an issue here for general use cases. You can also contact Mosaics Support [here](help.mosaics.ai).\n\n## Requirements\n\nPython 3.8 or above is required.\n\n## Documentation\n\nFor the latest documentation, see\n\n- [Mosaics AI](https://www.mosaics.ai)\n\n## Quickstart\n\nInstall the library with `pip install pymsx`\n\nNote: Don\'t hard-code authentication secrets into your Python. Use environment variables\n\nemail/Password Authentication:\n\n```bash\nexport MSX_USERNAME=*************\nexport MSX_PASSWORD=*************\n```\n\nIf you already have a token, use that instead:\n\n```bash\nexport MSX_TOKEN=*****************************************\n```\n\nExample usage:\n```python\nimport os\nimport pandas as pd\nfrom pymsx.client import MsxClient\n\n# If no credentials are supplied, then environment variables are required.\nemail = "help@mosaics.ai"\npassword = "$mosaics123"\n\n# ...or try using an active token.\n# This may fail, see exception handling below.\ntoken = None\n\n# First create client with active token or credentials\nmsx = MsxClient(\n    # ...using email/password\n    email=email,\n    password=password,\n    # ...or if using token, token will take priority\n    token=token\n)\n\n# Check the health of your server\nhealth = msx.health().dict()\n\nprint("Health: ", health)\n\nassert health is not None and health[\'status\'] == \'live\'\n\n# Add a dataset to your msx system\n\n# From a DataFrame\npath = "/path/to/dataset/data.csv"\ndf = pd.DataFrame(path)\nresult = msx.datasets.add(df=df)\n\n# Or pass in a string path to read from fs directly\nresult = msx.datasets.add(path=path)\n\nif result.ok:\n    print("DataFrame uploaded: ", result.details)\nelse\n    print("Upload failed: ", result.error)\n```\n\nException handling:\n```python\nfrom pymsx.client import MsxClient\nfrom pymsx.exceptions import ApiResponseError, InvalidTokenError\n\ntry:\n    try:\n        # An InvalidToken error is raised if the token is expired or incorrect\n        msx = MsxClient(\n            token=token\n        )\n    except InvalidTokenError:\n        print(f"Token invalid, logging in instead.")\n        # Catch all other errors using ApiResponseErrors\n        msx = MsxClient(\n            email=email,\n            password=password\n        )\nexcept ApiResponseError as e:\n    print(f"Could not create msx client: {e.error}")\n    return\n```\n\n## Contributing\n\nWe will allow contributing soon!\n\n## License\n\n[Apache License 2.0](LICENSE)\n',
    'author': 'mosaics.ai',
    'author_email': 'info@mosaics.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.mosaics.ai/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<3.10',
}


setup(**setup_kwargs)



"""Schemas for api responses and requests."""
from dataclasses import asdict, dataclass


@dataclass
class Credentials:
    """Credential model for requesting token."""

    email: str
    password: str

    dict = asdict


@dataclass
class TokenResponse:
    """Response when requesting api token."""

    token: str
    orgId: str

    dict = asdict


@dataclass
class ApiMessage:
    """Generic api response on certain routes."""

    message: str

    dict = asdict


@dataclass
class ApiError:
    """Generic api error response."""

    error: str

    dict = asdict


@dataclass
class HealthStatus:
    """Schema for remote server health status."""

    build: str
    sid: int
    status: str
    time: str

    dict = asdict


@dataclass
class CommandArgs:
    """Schema for issuing msx commands."""

    args: list[str]

    dict = asdict



"""Exceptions thrown by classes in pymsx."""
from typing import Optional

from pymsx.schemas import ApiError


class ApiResponseError(Exception):
    """Exception raised for api errors.

    Attributes:
        error: str - the error message if any
    """

    def __init__(self, error: Optional[ApiError] = None):
        """Raise api response error."""
        self.error = error.error if error is not None else "Unknown error"
        super().__init__(self.error)


class InvalidTokenError(Exception):
    """Exception raised when token invalid."""

    def __init__(self):
        """Raise invalid token error."""
        super().__init__("You are not authorized to make this request: Invalid token")
